LABEL_STUDIO_API_KEY = """API Key to authenticate to the Label Studio server. Can also be set with the LABELR_LABEL_STUDIO_API_KEY environment variable."""
LABEL_STUDIO_URL = """URL of the Label Studio server. Can also be set with the LABELR_LABEL_STUDIO_URL environment variable."""
LABEL_STUDIO_PROJECT_ID = "ID of the Label Studio project"
LABEL_STUDIO_VIEW_ID = "ID of the Label Studio view, used to filter tasks. If not provided, all tasks in the project are processed."
